from ._version import __version__  # noqa: F401
from .base import MedicalImage  # noqa: F401
# import Dicom to alias to the medimage namespace it here as well
from fileformats.application import Dicom  # noqa: F401
from .misc import (  # noqa: F401
    Analyze,
    Mgh,
    MghGz,
)
from .nifti import (  # noqa: F401
    Nifti,
    Nifti1,
    Nifti2,
    NiftiGz,
    NiftiX,
    NiftiGzX,
)
from .diffusion import (  # noqa: F401
    DwiEncoding,
    Bvec,
    Bval,
    NiftiBvec,
    NiftiGzBvec,
    NiftiXBvec,
    NiftiGzXBvec,
)
from .dicom import (  # noqa: F401
    DicomCollection,
    DicomDir,
    DicomSeries,
    # SiemensDicomDir,
)
from .raw import (  # noqa: F401
    Kspace,
    Rda,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
)
from .surface import (
    Gifti  # noqa: F401
)
from .contents.imaging.modality import (  # noqa: F401
    ImagingModality,
    CombinedModalities,
    DualEnergyXrayAbsorptiometry,
    Fluoroscopy,
    MrFluoroscopy,
    RadioFluoroscopy,
    MagneticResonanceImaging,
    DiffusionTensorImaging,
    DynamicContrast,
    EnhancedMagneticResonanceImaging,
    FunctionalMagneticResonanceImaging,
    MagneticResonanceAngiography,
    MagneticResonanceSpectroscopy,
    NuclearMedicineImaging,
    PositronEmissionTomography,
    PanographicRadiograph,
    ProjectionRadiography,
    ComputedRadiography,
    DigitalRadiography,
    DualEnergySubtractionRadiograpgy,
    Mammography,
    ScreenFilmRadiography,
    Stereoscopy,
    StereotacticRadiography,
    Spectroscopy,
    Tomography,
    ComputedTomography,
    Ultrasound,
    MRI,
    PET,
    CR,
    CT,
    DX,
    MG,
    MR,
    NM,
    PT,
    PX,
    RF,
    RG,
    US,
)
from .contents.imaging.derivatives import (
    Derivative,
    Mask
)
from .contents.anatomical_entity.material_anatomical_entity.anatomical_structure import (
    Brain, SpinalCord
)


import pytest
from medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c import get_image as get_dicom
from medimages4tests.dummy.nifti import get_image as get_nifti
from fileformats.core.exceptions import FormatMismatchError
from fileformats.medimage import Nifti1


def test_nifti_identify():
    Nifti1(get_nifti())


def test_nifti_not_identify():
    with pytest.raises(FormatMismatchError, match="No matching files with extensions"):
        Nifti1(get_dicom())



from . import MaterialAnatomicalEntity


class AnatomicalStructure(MaterialAnatomicalEntity):
    pass


class CardinalOrganPart(AnatomicalStructure):
    pass


class OrganRegion(CardinalOrganPart):
    pass


class OrganSegment(OrganRegion):
    pass


class SegmentOfNeuraxis(OrganRegion):
    pass


class Brain(SegmentOfNeuraxis):
    pass


class SpinalCord(SegmentOfNeuraxis):
    pass



from .base import (
    PetRawData,
    PetListMode,
    PetSinogram,
    PetCountRate,
    PetNormalisation,
)
from .siemens import (
    Vnd_Siemens_Biograph128Vision_Vr20b_PetRawData,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetListMode,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetSinogram,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetCountRate,
    Vnd_Siemens_Biograph128Vision_Vr20b_PetNormalisation,
)



from pathlib import Path
import typing as ty
from tempfile import mkdtemp
from random import Random
import pydicom
import numpy as np
from fileformats.core import FileSet
from fileformats.core.utils import SampleFileGenerator
from fileformats.medimage import MedicalImage, DicomCollection, DicomDir, DicomSeries
import medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c


@MedicalImage.read_array.register
def dicom_read_array(collection: DicomCollection) -> np.ndarray:
    image_stack = []
    for dcm_file in collection.contents:
        image_stack.append(pydicom.dcmread(dcm_file).pixel_array)
    return np.asarray(image_stack)


@MedicalImage.vox_sizes.register
def dicom_vox_sizes(collection: DicomCollection) -> ty.Tuple[float, float, float]:
    return tuple(
        collection.metadata["PixelSpacing"] + [collection.metadata["SliceThickness"]]
    )


@MedicalImage.dims.register
def dicom_dims(collection: DicomCollection) -> ty.Tuple[int, int, int]:
    return tuple(
        (
            collection.metadata["Rows"],
            collection.metadata["DataColumns"],
            len(list(collection.contents)),
        ),
    )


@DicomCollection.series_number.register
def dicom_series_number(collection: DicomCollection) -> str:
    return str(collection.metadata["SeriesNumber"])


@FileSet.generate_sample_data.register
def dicom_dir_generate_sample_data(
    dcmdir: DicomDir,
    generator: SampleFileGenerator,
) -> ty.Iterable[Path]:
    dcm_dir = medimages4tests.dummy.dicom.mri.t1w.siemens.skyra.syngo_d13c.get_image()
    series_number = generator.rng.randint(1, SERIES_NUMBER_RANGE)
    dest = generator.generate_fspath(DicomDir)
    dest.mkdir()
    for dcm_file in dcm_dir.iterdir():
        dcm = pydicom.dcmread(dcm_file)
        dcm.SeriesNumber = series_number
        pydicom.dcmwrite(dest / dcm_file.name, dcm)
    return [dest]


@FileSet.generate_sample_data.register
def dicom_series_generate_sample_data(
    dcm_series: DicomSeries,
    generator: SampleFileGenerator,
) -> ty.Iterable[Path]:
    dicom_dir: Path = dicom_dir_generate_sample_data(dcm_series, generator=generator)[0]
    stem = generator.generate_fspath().stem
    fspaths = []
    for i, dicom_file in enumerate(dicom_dir.iterdir(), start=1):
        fspaths.append(dicom_file.rename(generator.dest_dir / f"{stem}-{i}.dcm"))
    dicom_dir.rmdir()
    return fspaths


SERIES_NUMBER_TAG = ("0020", "0011")
SERIES_NUMBER_RANGE = int(1e8)



import pytest
from fileformats.medimage import DicomSeries, DicomDir
from conftest import OLD_MRTRIX_VERSION


@pytest.mark.xfail(condition=OLD_MRTRIX_VERSION, reason="Old MRtrix version")
def test_dicom_series_metadata(tmp_path):
    series = DicomSeries.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series.metadata["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series.metadata["SOPInstanceUID"], list)


@pytest.mark.xfail(condition=OLD_MRTRIX_VERSION, reason="Old MRtrix version")
def test_dicom_dir_metadata(tmp_path):
    series = DicomDir.sample(tmp_path)

    # Check series number is not a list
    assert not isinstance(series.metadata["SeriesNumber"], list)
    # check the SOP Instance ID has been converted into a list
    assert isinstance(series.metadata["SOPInstanceUID"], list)



import typing as ty
from operator import itemgetter
from collections import defaultdict, Counter
from pathlib import Path
from functools import cached_property
from fileformats.core import hook, FileSet
from fileformats.generic import DirectoryContaining, SetOf, TypedSet
from fileformats.application import Dicom
from .base import MedicalImage

# =====================================================================
# Custom loader functions for different image types
# =====================================================================


class DicomCollection(MedicalImage):
    """Base class for collections of DICOM files, which can either be stored within a
    directory (DicomDir) or presented as a flat list (DicomSeries)
    """

    content_types = (Dicom,)
    iana_mime = None

    def __len__(self):
        return len(self.contents)

    @hook.extra
    def series_number(self) -> str:
        raise NotImplementedError

    @cached_property
    def contents(self) -> ty.List[Dicom]:
        return sorted(super().contents, key=lambda d: d.metadata["SOPInstanceUID"])


class DicomDir(DicomCollection, DirectoryContaining[Dicom]):
    pass


class DicomSeries(DicomCollection, SetOf[Dicom]):
    @classmethod
    def from_paths(
        cls, fspaths: ty.Iterable[Path], common_ok: bool = False,
        selected_keys: ty.Optional[ty.Sequence[str]] = None
    ) -> ty.Tuple[ty.Set["DicomSeries"], ty.Set[Path]]:
        """Separates a list of DICOM files into separate series from the file-system
        paths

        Parameters
        ----------
        fspaths : ty.Iterable[Path]
            the fspaths pointing to the DICOM files
        common_ok : bool, optional
            included to match the signature of the overriden method, but ignored as each
            dicom should belong to only one series.
        selected_keys : ty.Optional[ty.Sequence[str]], optional
            metadata keys to load from the DICOM files, typically used for performance
            reasons, by default None (i.e. all metadata is loaded)

        Returns
        -------
        tuple[set[DicomSeries], set[Path]]
            the found dicom series objects and any unrecognised file paths
        """
        dicoms, remaining = Dicom.from_paths(fspaths, common_ok=common_ok)
        series_dict = defaultdict(list)
        for dicom in dicoms:
            dicom.select_metadata(selected_keys)
            series_dict[
                (str(dicom.metadata["StudyInstanceUID"]), str(dicom.metadata["SeriesNumber"]))
            ].append(dicom)
        return set([cls(s) for s in series_dict.values()]), remaining


@FileSet.read_metadata.register
def dicom_collection_read_metadata(
    collection: DicomCollection, selected_keys: ty.Optional[ty.Sequence[str]] = None
) -> ty.Mapping[str, ty.Any]:
    # Collated DICOM headers across series
    collated = {}
    key_repeats = Counter()
    varying_keys = set()
    # We use the "contents" property implementation in TypeSet instead of the overload
    # in DicomCollection because we don't want the metadata to be read ahead of the
    # the `select_metadata` call below
    base_class = TypedSet if isinstance(collection, DicomSeries) else DirectoryContaining
    for dicom in base_class.contents.fget(collection):
        dicom.select_metadata(selected_keys)
        for key, val in dicom.metadata.items():
            try:
                prev_val = collated[key]
            except KeyError:
                collated[key] = val  # Insert initial value (should only happen on first iter)
                key_repeats.update([key])
            else:
                if key in varying_keys:
                    collated[key].append(val)
                # Check whether the value is the same as the values in the previous
                # images in the series
                elif val != prev_val:
                    collated[key] = [prev_val] * key_repeats[key] + [val]
                    varying_keys.add(key)
                else:
                    key_repeats.update([key])
    return collated



from fileformats.application import Gzip
from fileformats.generic import File
from fileformats.core import hook
from fileformats.core.mixin import WithSeparateHeader, WithMagicVersion
from .base import MedicalImage


# ==================
# Other Data Formats
# ==================


class AnalyzeHeader(File):

    ext = ".hdr"


class Analyze(WithSeparateHeader, MedicalImage, File):

    ext = ".img"
    header_type = AnalyzeHeader


class Mgh(WithMagicVersion, File):
    """
    FreeSurfer 4-dimensional brain images

    See https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
    """

    ext = ".mgh"
    magic_pattern = rb"(....)"  # First integer is the version string

    @hook.check
    def is_supported_version(self):
        self.version == 1


class MghGz(Gzip[Mgh]):
    """
    FreeSurfer 4-dimensional brain images, gzipped

    See https://surfer.nmr.mgh.harvard.edu/fswiki/FsTutorial/MghFormat
    """

    iana_mime = "application/x-mgh+zip"
    ext = ".mgz"
